import java.util.ArrayList;

import ilog.concert.IloException;
import ilog.concert.IloLinearNumExpr;
import ilog.concert.IloNumVar;
import ilog.cplex.IloCplex;

public class Algorithms_BinPacking {
	static ArrayList<Double> fillLevels;
	static String[] statuses; //"unpacked", "packed"
	
	static int solveInstance(String algorithm, double gradualLossFactor, int maxGradualHorizon, Instance_BinPacking instance, int lookahead){
		fillLevels = new ArrayList<Double>(); 
		int T = instance.T;
		statuses = new String[T];
		for(int t = 0; t < T; t++)
			statuses[t] = "unpacked";
		for(int t = 0; t < T; t++)
			packItem(algorithm, gradualLossFactor, maxGradualHorizon, instance, lookahead, t);				
		return fillLevels.size();
	}
	
	static void packItem(String algorithm, double gradualLossFactor, int maxGradualHorizon, Instance_BinPacking instance, int lookahead, int t_current){
		int indexOfNextItem = getIndexOfLargestItem(instance, lookahead, t_current);
		int indexOfBin = -1;
		if(algorithm.equals("BFD"))
			indexOfBin = getIndexOfBin_BFD(instance.sizes[indexOfNextItem]);
		if(algorithm.equals("SnapshotMIP"))
			indexOfBin = getIndexOfBin_SnapshotMIP(false, gradualLossFactor, maxGradualHorizon, indexOfNextItem, instance, lookahead, t_current);
		if(algorithm.equals("SnapshotMIPForecasts"))
			indexOfBin = getMostFrequentIndexOfBin(gradualLossFactor, maxGradualHorizon, indexOfNextItem, instance, lookahead, t_current);
		if(indexOfBin != -1)
			fillLevels.set(indexOfBin, fillLevels.get(indexOfBin) + instance.sizes[indexOfNextItem]);
		else
			fillLevels.add(instance.sizes[indexOfNextItem]);
		statuses[indexOfNextItem] = "packed";
	}
	
	static int getMostFrequentIndexOfBin(double gradualLossFactor, int maxGradualHorizon, int indexOfNextItem, Instance_BinPacking instance, int lookahead, int t_current){
		int numberOfShootings = 25;
		int[] indizesOfBin = new int[numberOfShootings];
		for(int it = 0; it < numberOfShootings; it++)
			indizesOfBin[it] = getIndexOfBin_SnapshotMIP(true, gradualLossFactor, maxGradualHorizon, indexOfNextItem, instance, lookahead, t_current);
		return (new Distribution(indizesOfBin)).getValueWithHighestProbability();
	}
	
	static int getIndexOfLargestItem(Instance_BinPacking instance, int lookahead, int t_current){
		int index = -1;
		double largestSize = -1;
		for(int t = 0; t < Math.min(t_current + lookahead, instance.T); t++)
			if(statuses[t].equals("unpacked") && instance.sizes[t] > largestSize){
				index = t;
				largestSize = instance.sizes[t];
			}
		return index;		
	}
	
	static int getIndexOfBin_BFD(double size){
		int index = -1;
		double largestFillLevel = -1;
		for(int i = 0; i < fillLevels.size(); i++)
			if(fillLevels.get(i) + size <= 1 && fillLevels.get(i) > largestFillLevel){
				largestFillLevel = fillLevels.get(i);
				index = i;
			}
		return index;
	}
	
	static int getIndexOfBin_SnapshotMIP(boolean forecasts, double gradualLossFactor, int maxGradualHorizon, int indexOfItem, Instance_BinPacking instance, int lookahead, int t_current){
		int i_selected = -1;
		int n;
		if(!forecasts)
			n = Math.min(lookahead, instance.T - t_current);
		else
			n = Math.min(instance.T - t_current, lookahead + maxGradualHorizon);
		double s[] = new double[n];
		int current_i = 0;
		for(int t = 0; t < Math.min(t_current + lookahead, instance.T); t++){
			if(statuses[t].equals("unpacked") && current_i < n){
				s[current_i] = instance.sizes[t];
				if(t == indexOfItem)
					i_selected = current_i;
				current_i++;
			}
		}
		if(forecasts){
			instance.createForecasts(gradualLossFactor, t_current, lookahead);
			for(int t = Math.min(t_current + lookahead, instance.T); t < instance.T; t++)
				if(current_i < n){
					s[current_i] = instance.sizes_forecasted[t];
					current_i++;
				}
		}
		int indexOfBin = -1;
		try{
			IloCplex cplex = new IloCplex();
			
			IloNumVar[][] x = new IloNumVar[n][];
			for(int i = 0; i < n; i++)
				x[i] = cplex.boolVarArray(n);
			IloNumVar[][] x_old = new IloNumVar[n][];
			for(int i = 0; i < n; i++)
				x_old[i] = cplex.boolVarArray(fillLevels.size());
			
			IloNumVar[] y = cplex.boolVarArray(n);		
			
			IloLinearNumExpr objective = cplex.linearNumExpr();
			for(int i = 0; i < n; i++)
				objective.addTerm(1, y[i]);
	        cplex.addMinimize(objective);
	        
	        for(int i = 0; i < n; i++){
	        	IloLinearNumExpr expr = cplex.linearNumExpr();
	        	for(int j = 0; j < n; j++)
	        		expr.addTerm(1, x[i][j]);
	        	for(int j = 0; j < fillLevels.size(); j++)
	        		expr.addTerm(1, x_old[i][j]);
	        	cplex.addEq(expr, 1);
	        }
	        for(int j = 0; j < n; j++){
	        	IloLinearNumExpr expr = cplex.linearNumExpr();
	        	for(int i = 0; i < n; i++)
	        		expr.addTerm(s[i], x[i][j]);
	        	expr.addTerm(-1, y[j]);
	        	cplex.addLe(expr, 0);
	        }
	        for(int j = 0; j < fillLevels.size(); j++){
	        	IloLinearNumExpr expr = cplex.linearNumExpr();
	        	for(int i = 0; i < n; i++)
	        		expr.addTerm(s[i], x_old[i][j]);
	        	cplex.addLe(expr, 1 - fillLevels.get(j));
	        }
	        cplex.setOut(null);
	        cplex.setParam(IloCplex.DoubleParam.EpGap, 0.05);
	        cplex.setParam(IloCplex.DoubleParam.TiLim, 60);
	        cplex.solve();
	        for(int j = 0; j < fillLevels.size(); j++)
	        	if(Math.abs(cplex.getValue(x_old[i_selected][j]) - 1) < 0.001)
	        		indexOfBin = j;
	        cplex.end();			
		}
	    catch (IloException ex){
	    	System.out.println("Concert Error: " + ex);
	    }
		return indexOfBin;
	}	
}